/*****************************************************************************
 *
 * NAME
 *    ldp_cmds.h (renamed SonyLDPcodes.h by Eric Celeste 930210)
 *
 * CONTENTS
 *    This file contains some the commands that can be sent to
 *    the Sony 1450 LDP
 *
 *****************************************************************************/


#ifndef _LDP_CMDS_H
#define _LDP_CMDS_H

/*********************** codes returned by the ldp *********************
 * ACK
 *    returned from the player when a valid command has been recieved and
 *    when the player is ready for the next command.  This means the code
 *    is returned twice to a single command excpet for the MOTOR_ON
 *    and MOTOR_OFF
 *
 * COMPLETION
 *    is sent when the ldp is finished witha a search, repeat, or m-search
 *
 * NACK
 *    returned when the LDP isn't happy
 *
 * ERROR
 *    returned if there is a RS-232 protocol error
 *
 * LID_OPEN
 *    returned if the lid is open to some commands
 *
 * MARK_RETURN
 *    returned when a mark position is read in by the MARK_SET cmd
 *
 * NO_FRAME
 *    returned when trying to reach a non-existant frame
 ***/
#define LDP_ACK		0x0A
#define LDP_NACK	0x0B
#define LDP_COMPLETION  0x01
#define LDP_ERROR       0x02
#define LDP_LID_OPEN    0x03
#define LDP_NOT_TARGET  0x05
#define LDP_NO_FRAME    0x06

/********************** codes sent to the ldp **********************
 * ADDR_INQ
 *    inquires about the current address.  For CAV, this is a frame
 *    number, 5 bytes -- the 1st byte is the most sig digit
 *
 * CHAPTER MODE
 *    sets the ldp to chapter indexing mode
 *
 * CHAPTER_INQ
 *   returns in a 2-byte ascii code, the 1st byte is the msb
 *
 * AUDIO_MUTE_OFF
 * AUDIO_MUTE_ON
 *    this will only have effect if the disk is moving
 *
 * CH1_ON
 * CH2_ON
 * CH1_OFF
 * CH2_OFF
 *    these have effect regardless of the current state of the ldp
 *
 * RESET
 *    also referred to as CLEAR_ALL, this takes the ldp out of search\
 *    or repeat mode and puts it into still mode
 *
 * CX_ON
 * CX_OFF
 *    these commands affect the CX noise reduction system, which
 *    greatly improves the dynamic range of the audio
 *
 * EJECT
 *    ejects the disks.  this causes a feedback of LID_OPEN if there
 *    is a disk, else a NAK
 *
 * EJECT_ENABLE
 * EJECT_DISABLE
 *    these affect the eject button on the front of the panel
 *
 * ENTER
 *    this is used as an end marker for the input of parameters for 
 *    various commands, such as SEARCH, REPEAT, STEP, and MARK_SET
 *
 * SCAN_F
 * SCAN_R
 *    play the disk at scan speed (100x)
 *
 * FAST_F
 * FAST_R
 *    move the disk at high speed (3x)
 *
 * PLAY_F
 * PLAY_R
 *    play the disk at normal speed
 *
 * SLOW_F
 * SLOW_R
 *   play the disk at slow speed (.2x)
 * 
 * STEP_F
 * STEP_R
 *   play the disk at variable speed (1 - 1/255)
 * 
 * ONE_FRAME_F
 * ONE_FRAME_R
 *    this is also referred to as "step and still"
 *
 * FRAME_MODE
 *     sets the ldp to frame mode
 *
 * INDEX_ON
 * INDEX_OFF
 *     turns on or off the on-screen index
 *
 * MARK_SET
 *    causes the ldp to spit out a MARK_RETURN when the set position is
 *    read
 *
 * M-SEARCH
 * MEMORY
 *    MEMORY memorizes the current frame, while M-SEARCH searches
 *    to the frame last stored in MEMORY
 *
 * REPEAT
 *    allows the designated sequence to be repeated a number of times
 *    at a designated speed.  see manual for more details.
 *
 * SEARCH
 *    locateds the target address
 *
 * STATUS_INQ
 *    hey, laser-disc-dude, what's up? (btw, this returns 5 bytes of data
 *
 * STILL
 * STOP
 *    self-explanitory, right?
 *
 ***/

/* modes */
#define LDP_CHAPTER_MODE 0x69
#define LDP_CHAPTER_INQ 0x76
#define LDP_ADDR_INQ 	0x60
#define LDP_FRAME_MODE 0x55

/* audio */
#define LDP_AUDIO_MUTE_OFF 0x25
#define LDP_AUDIO_MUTE_ON  0x24
#define LDP_CH1_ON      0x46
#define LDP_CH1_OFF     0x47
#define LDP_CH2_ON      0x48
#define LDP_CH2_OFF     0x49

/* noise reduction */
#define LDP_CX_ON       0x6E
#define LDP_CX_OFF      0x6F

/* resetting */
#define LDP_RESET	0x56
#define LDP_CLEAR_ALL   LDP_RESET

/* ejecting */
#define LDP_EJECT       0x2A
#define LDP_EJECT_ENABLE 0x74
#define LDP_EJECT_DISABLE 0x75

/* playing the disk */
#define LDP_SCAN_F      0x3E
#define LDP_SCAN_R      0x4E
#define LDP_FAST_F	0x3B
#define LDP_FAST_R	0X4B
#define LDP_PLAY_F	0x3A
#define LDP_PLAY_R	0X4A
#define LDP_SLOW_F	0x3C
#define LDP_SLOW_R	0X4C
#define LDP_STEP_F	0X3D
#define LDP_STEP_R	0X4D
#define LDP_ONE_FRAME_F 0x2B
#define LDP_ONE_FRAME_R 0x2C

#define LDP_INDEX_ON    0x50
#define LDP_INDEX_OFF   0x51

#define LDP_MEMORY   0x5A
#define LDP_M_SEARCH 0x5B

#define LDP_MOTOR_ON 0x62
#define LDP_MOTOR_OFF 0x63

#define LDP_STOP	0X3F
#define LDP_STILL	0X4F
#define LDP_SEARCH  	0x43

#define LDP_ENTER	0x40
#define LDP_REPEAT	0x44

#define LDP_STATUS_INQ     0x67

#endif


